/* -- File Info --
 * generic.cpp - Implementation file
 *
 * Several generic functions for Orgestrate and Orgestrate OrgPlay.
 * 
 *   ~nmn
 */

/* -- Legal Info --
 * This code is Public Domain.
 */

#ifndef GENERIC_CPP
#define GENERIC_CPP

class RectangleShape
{
	public:
		double X, Y, W, H;
		
		RectangleShape() { X=0; Y=0; W=0; H=0; }
		RectangleShape(double p_W, double p_H) { X=0; Y=0; W=p_W; H=p_H; }
		RectangleShape(double p_X, double p_Y, double p_W, double p_H) { X=p_X; Y=p_Y; W=p_W; H=p_H; }
		
		void Translate(double p_X, double p_Y) { X+=p_X; Y+=p_Y; }
		void Scale(double p_X, double p_Y) { X*=p_X; Y*=p_Y; W*=p_X; H*=p_Y; }
		
	private:
};

class Position
{
	public:
		double X, Y;
		
		Position() { X=0; Y=0; }
		Position(double p_X, double p_Y) { X=p_X; Y=p_Y; }
		
		void Translate(double p_X, double p_Y) { X+=p_X; Y+=p_Y; }
		void Rotate(double p_Rotate) { X=cos(p_Rotate)*X; Y=sin(p_Rotate)*Y; }
		void Scale(double p_X, double p_Y) { X*=p_X; Y*=p_Y; }
		
	private:
};

/***! Microsoft Windows Specific Code Below !***/
// Because Microsoft sucks. POSIX works, to say the least.
// Even Unix hating programmers can't stand the Win32 API.
#ifdef __WIN32__
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// void usleep(unsigned long long length)
// --
// Pauses the application for a moment on Win32.
// --
// p_Length - Length in nanoseconds to sleep for
static inline void usleep(unsigned long long p_Length)
{
	// Sleep using Microsoft Sleep. Measures in Milliseconds as opposed to POSIX nanoseconds.
	Sleep(p_Length/1000);
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// std::string toString(int p_Number)
// --
// Converts p_Number to an std::string.
// --
// p_Number: The number to be converted.
static inline std::string toString(int p_Number)
{
	// Local Variables
	char l_NumAsStr[255];
	
	// Use the standard "sprintf" to print p_Number to l_NumAsStr using formatted strings.
	// Notice that this operation can get pretty slow on certain OS's / Computers.
	sprintf(l_NumAsStr, "%d", p_Number);

	// Return l_NumAsStr in string form.
	return string(l_NumAsStr);
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FmodErrorCheck(FMOD_RESULT p_Result)
// --
// Checks the result for error.
// --
// p_Result: Result to check for errors.
static inline void FmodErrorCheck(FMOD_RESULT p_Result)
{
	// Check and see if there was an error.
	if (p_Result != FMOD_OK)
	{
		
		// If there is an error, print it out.
		printf("Audio error [No. %d] \n--\n%s\n--\n\n", p_Result, FMOD_ErrorString(p_Result));
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// InitFMOD()
// --
// Initalizes an FMOD device
static inline void InitFMOD()
{
	// Local Variables
	FMOD_RESULT l_Result;
	FMOD_DSP *l_Effect;
	unsigned int l_Version;
	
	// Create a blank FMOD System
	l_Result = FMOD_System_Create(&Org::Device);
	FmodErrorCheck(l_Result);

	// Get the dynamic library version number
	l_Result = FMOD_System_GetVersion(Org::Device, &l_Version);			      
	FmodErrorCheck(l_Result);
	
	// Check vs. header version
	// Does not cause a fatal error but issues a warning.
	if (l_Version < FMOD_VERSION)
	{
		// Print out the warning message.
		printf("LIBRARY VERSION MISMATCH!\n--\nlibfmodex: %08x\nfmod.h: %08x\n--\nThe FMOD version installed must be the same or as new as the headers used to compile the\napplication or else the application will fail to execute properly.\n\nContinuing in five seconds. Press Ctrl+C to exit.\n\n", l_Version, FMOD_VERSION);
		
		// Give the user time to notice the message.
		usleep(5000000);
		
		// Return nothing to the user.
		return;
	}
	
	// ALSA has the ability to run multiple applications at once. Because it is much lower latency and usually more capable than
	// any sound daemon or other sound libraries on Linux, it will be prefered when possible.
	// 
	// This is default because now DMIX is default for ALSA cards needing it, meaning multiple applications will work by default!
	// 
	// Notice: Only one OSS application can run at a time, even if with other ALSA apps via emulation.
	// This is because of a design flaw and there happens to be no way around it.
	#ifdef __linux__
	l_Result = FMOD_System_SetOutput(Org::Device, FMOD_OUTPUTTYPE_ALSA);
	FmodErrorCheck(l_Result);
	#endif
	
	// Set the format of the output. Its generally OK to use 44100 but ORGestrate is meant to be general purpose. 22050
	// does well for ORG music considering its not all that HQ anyways, and for most games its OK also. Its also quite fast.
	l_Result = FMOD_System_SetSoftwareFormat(Org::Device, 48000, FMOD_SOUND_FORMAT_PCM16, 2, 0, FMOD_DSP_RESAMPLER_LINEAR);
	FmodErrorCheck(l_Result);
	
	l_Result = FMOD_System_SetSpeakerMode( Org::Device, FMOD_SPEAKERMODE_STEREO );
	FmodErrorCheck(l_Result);
	
	// Initalize FMOD with 32 accelerated channels. ORG only has 16 channels that cannot contain overlapping notes
	// but if an application wishes to play more sounds (UI sounds, or game sounds even) we should atleast reserve a few.
	// 
	// Plus ORGestrate Sampler is neutral of the song playing. More than one could play or other things can be done with it.
	l_Result = FMOD_System_Init(Org::Device, 32, FMOD_INIT_NORMAL, NULL);
	FmodErrorCheck(l_Result);
	
	l_Result = FMOD_System_GetMasterChannelGroup(Org::Device, &Org::MasterCG);
	FmodErrorCheck(l_Result);
	
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FMOD_SOUND *FMOD_LoadSample(string p_Filename, int p_Loop = 0)
// --
// Loads a sample into FMOD and returns it.
// --
// p_Filename: Name of the file to open.
// p_Loop: 1=yes, 0=no.
static inline FMOD_SOUND *FMOD_LoadSample(string p_Filename, int p_Loop = 0)
{
	// Local Variables
	FMOD_SOUND *l_Sound = 0;
	FMOD_RESULT l_Result;
	
	// Strange problems with Microsoft Windows + FMOD may cause FMOD to randomly die before sounds even begin to get loaded.
	if(!Org::Device)
	{
		fprintf(stderr,"FMOD hiccuped. Expect problems with samples missing.");
		
		// This does sometimes occur. Just recreate the device.
		InitFMOD();
	}
	
	// Create the sound itself. Make sure that went OK.
	if(p_Loop)
	{
		l_Result = FMOD_System_CreateSound(Org::Device, p_Filename.c_str(), FMOD_SOFTWARE | FMOD_LOOP_NORMAL, 0, &l_Sound);
	}
	else
	{
		l_Result = FMOD_System_CreateSound(Org::Device, p_Filename.c_str(), FMOD_SOFTWARE | FMOD_LOOP_OFF, 0, &l_Sound);
	}
	FmodErrorCheck(l_Result);
	
	// Return l_Sound to the calling function.
	return l_Sound;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FMOD_CHANNEL *FMOD_PlaySound(FMOD_SOUND *p_Sound, unsigned int p_Frequency, float p_Panning, unsigned int p_Volume)
// --
// Plays a sample previously loaded to FMOD.
// --
// p_Sound: Sound to play.
// p_Frequency: Frequency to play it at.
// p_Panning: Panning of the sound.
// p_Volume: Volume of the sound.
static inline FMOD_CHANNEL *FMOD_PlaySound(FMOD_SOUND *p_Sound, unsigned int p_Frequency, float p_Panning, unsigned int p_Volume, FMOD_CHANNEL *p_Channel, bool p_Reserve)
{
	// Local Variables
	FMOD_RESULT l_Result;
	FMOD_CHANNEL *l_Channel = p_Channel;
	bool l_FailTwo = false;
	
	// Strange problems with Microsoft Windows + FMOD may cause FMOD to randomly die before sounds even begin to get loaded.
	if(!Org::Device)
	{
		// This does sometimes occur. Just recreate the device.
		InitFMOD();
	}
	
	// Play the sound. Stuff the channel into l_Channel.
	if( !(p_Channel && p_Reserve) )
	{
tryplay:
		l_Result = FMOD_System_PlaySound(Org::Device, FMOD_CHANNEL_FREE, p_Sound, 0, &l_Channel);
		FmodErrorCheck(l_Result);
	}
	
	// Change the volume.
	l_Result = FMOD_Channel_SetVolume(l_Channel, (float)(p_Volume+25.0)/255.0);
	if(l_Result != FMOD_OK && !l_FailTwo)
	{
		l_FailTwo = true;
		
		// TODO: Better code. This is ABSOLUTELY TERRIBLE.
		// Any other way to tell if the channel is valid?
		goto tryplay;
	}
	FmodErrorCheck(l_Result);
	
	// Change the panning.
	l_Result = FMOD_Channel_SetPan(l_Channel, p_Panning);
	FmodErrorCheck(l_Result);
	
	// Change the frequency
	l_Result = FMOD_Channel_SetFrequency(l_Channel, p_Frequency);
	FmodErrorCheck(l_Result);

	// Return l_Channel to the calling function.
	return l_Channel;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// void FMOD_StopSound(FMOD_CHANNEL** p_Channel)
// --
// Stops a sound previously played with FMOD.
// --
// p_Channel: Channel the sound is currently playing on.
static inline void FMOD_StopSound(FMOD_CHANNEL** p_Channel)
{
	// Stop the channel.
	FMOD_Channel_Stop(*p_Channel);
	
	// Discard the channel.
	*p_Channel = 0;
	
	// Return nothing.
	return;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif
