/* -- File Info --
 * org.h - Header file
 */

/* -- Legal Info --
 * This code is public domain.
 */

#ifndef ORG_H
#define ORG_H

#define PACK __attribute__((packed))

#include <string>
#include <vector>
#include <iostream>
#include <fstream>
#include <ctime>
#include <cmath>
#include <unistd.h>

using namespace std;

#ifdef WIN32

#include <windows.h>
#include <fmodex/fmod.h>
#include <fmodex/fmod_errors.h>
//#include <GL/glew.h>

#else

#include <fmodex/fmod.h>
#include <fmodex/fmod_errors.h>

#endif

#include <SDL/SDL.h>

namespace Org
{
	extern FMOD_SYSTEM *Device;
	extern FMOD_CHANNELGROUP *MasterCG;

	// Notice that ORGestrate will not support values other than this for editing.
	enum SPBABPS
	{
		_2_4,
		_3_3,
		_3_4,
		_4_3,
		_4_4,
		_6_3,
		_6_4,
		_8_4,
		_ARB	// Arbitrary, for when a file already contains a value other than the above.
	};

	// Header of the file.
	struct Header
	{
		unsigned short tempo;
		unsigned char steps_bar;
		unsigned char beats_step;
		unsigned int loopstart;
		unsigned int loopstop;
	} PACK;
	
	// Each instrument is a triangular wave?
	struct Instrument
	{
		unsigned short pitch;
		unsigned char id;	// 0-99 for sound, 0-11 for percussion
		unsigned char noloop;
		unsigned short numnotes;
	} PACK;
	
	struct SampleData
	{
		int note;	// What note is the sample playing on?
		int length;	// length it is supposed to play for.
		int volume;	// Volume it should play at
		float pan;	// ( ( -1 ||0|| 1 ) )
		int pitch;	// What is the pitch of the sample?
		int sample;	// Sample index
		int loop;	// Whether or not to loop. x?0:1
		int playing;	// Mode of the note. 0 if we hit a blank. 1 if first hit. 2 if loop hit.
		int instrument;	// insturment the sample comes from
	};

	// Sample playback engine
	class Sampler
	{
		private:
			// ---- Samples Variables --------------// Below are variables used for holding samples.
			FMOD_SOUND *melodysamples[100];		//  Melody Samples
			FMOD_SOUND *percussionsamples[11];	//  Percussion Samples
			// ---- Voice Variables ----------------// Below are the variables that control playback.
			int channellength[16];			//  The length of time the voice shall continue for.
			FMOD_SOUND *channelsample[16];		//  The coresponding samples of the channels
			FMOD_CHANNEL *channelindex[16];		//  Instruments in each voice.
			unsigned int pitch[12*8];		//  Pitches
		public:
			Sampler();
			void SetMasterVolume(float p_Volume);
			bool HitNote(SampleData notehit);
			bool Update(int ms);
	};
	
	// Song class
	class OrgSong
	{
		private:
			// ---- File Structures ----------------// Below are the blocks of data stored in the file.
			Header header;				//  Header of the entire file.
			vector<unsigned int> notepos[16];	//  Note positions.
			vector<unsigned char> notenote[16];	//  Note datas.
			vector<unsigned char> notelength[16];	//  Note datas.
			vector<unsigned char> notevolume[16];	//  Note datas.
			vector<unsigned char> notepan[16];	//  Note datas.
			// ---- Playback Variables -------------// Below are variables used for playback.
			long double cursor;			//  Cursor, in beats.
			bool isPlaying;				//  Whether the music is playing or not.
			bool useLoopPoints;			//  Whether to use the loop points or not.
			unsigned long msSinceBeat;		//  Milliseconds since last beat.
			unsigned long msLastUpdate;		//  Exact measure of milliseconds last Update() occured at.
		public:
			// ---- Constructors -------------------// The below are constructors.
			OrgSong();				//  Create a blank song.
			OrgSong(string filename);		//  Load a song from file.
			// ---- Loading ------------------------// The below handles loading songs.
			bool LoadBlank();			//  Makes a blank Org song.
			bool LoadFromFile(string filename);	//  Loads from a .ORG file.
			// ---- Saving -------------------------// The below is for saving an ORG back to file.
			bool SaveToFile(string filename);       //  Saves the currently loaded ORG file.
			// ---- Playback Control ---------------// Typically the below is only needed for playback.
			int  Update();				//  Update song cursor and other playback related features.
			bool IsPlaying();			//  Returns true if the song is playing.
			bool Play();				//  Starts/Continues playback. Returns isPlaying additionally.
			bool Play(unsigned int at);		//  Starts playback at point.
			bool Pause();				//  Stops playback without clearing cursor.
			bool Stop();				//  Stops playback and clears cursor.
			bool SetCursor(unsigned int to);	//  Sets the cursor to point.
			bool UsingLoopPoints();			//  Returns true if the song is set to use loop points
			bool UseLoopPoints(bool use);		//  Sets the loop point usage.
			unsigned int GetCursor();		//  Returns the cursor.
			// ---- Playback -----------------------// The below are for playback and synthisization.
			SampleData GetNoteHit(int instrument);	//  Checks for a note hit.
			SampleData GetNoteHit(int , int);	//  Checks for a note hit at point.
			// ---- Editing ------------------------// Typically the below is only needed for editing.
			unsigned short GetTempo();		//  Returns the songs tempo.
			bool SetTempo(unsigned short to);	//  Sets the songs tempo.
			SPBABPS GetSPBABPS();			//  Gets the Steps Per Bar And Beats Per Step.
			bool SetSPBABPS(SPBABPS to);		//  Sets the Steps Per Bar And Beats Per Step.
			unsigned int GetLoopStart();		//  Gets the starting loop point.
			unsigned int GetLoopStop();		//  Gets the stopping loop point.
			bool SetLoopStart(unsigned int to);	//  Sets the starting loop point to another beat.
			bool SetLoopStop(unsigned int to);	//  Sets the stopping loop point to another beat.
			
			bool cursorUpdated;
			Instrument instruments[16];		//  Global instrument data.
			unsigned long msPerBeat;		//  Calculation of ms/beat so we don't end up doing it per frame.
	};
}

#endif
