/* -- File Info --
 * orgGL.cpp - Implementation file
 *
 * This file is a simple GUI for use in ORGplay and other software. You may use it in your own projects if you wish.
 *
 * To note, this code is PD. You may copy it, redistribute it, and use it as you wish, unrestricted.
 *   ~nmn
 */

/* -- Legal Info --
 * This code is Public Domain.
 */

#include "orgGL.h"

bool GraphicsOpenGL::Init(int p_Width, int p_Height, int p_PreferedBPP)
{
	// Initalize the Video subsystem of SDL.
	if(SDL_InitSubSystem(SDL_INIT_VIDEO)) 
	{
		printf("Error.\n%s\n--\nThe graphics context has not been created.\n--\n\n", SDL_GetError());
		return false;
	}
	
	// Enable double buffering for SDL_GL
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
	
	// Create a new SDL_GL surface.
	screen = SDL_SetVideoMode(p_Width, p_Height, p_PreferedBPP, SDL_OPENGL);
	
	// Make sure the pointer is valid.
	if(!screen)
	{
		printf("Error.\nUnknown SDL/OpenGL failure.\n--\nThe graphics context has not been created.\n--\n\n");
		return false;
	}
	
	// Set the screen clear color to blackness.
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	
	// Make the viewport fill the drawing area.
	glViewport( 0, 0, p_Width, p_Height );
	
	// Setup an orthogonal drawing matrix that maps pixel-for-pixel
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(0.0f, (float)p_Width, (float)p_Height, 0.0f, -1.0f, 1.0f);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	// Enable texturing.
	glEnable(GL_TEXTURE);
	glEnable(GL_TEXTURE_2D);
	
	// Do the inital screen clearing
	glClear(GL_COLOR_BUFFER_BIT);
	
	// Not useful. Just makes absolute sure the depth buffer doesn't attack.
	glDepthFunc(GL_ALWAYS);
	
	// Enable blending.
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	fontlarge = new FTGLTextureFont("gfx/FreeSans.ttf");
	fontsmall = new FTGLTextureFont("gfx/FreeSans.ttf");
	
	fontlarge->FaceSize(24);
	fontlarge->Depth(20);
	fontlarge->CharMap(ft_encoding_unicode);
	
	fontsmall->FaceSize(12);
	fontsmall->Depth(20);
	fontsmall->CharMap(ft_encoding_unicode);
	
	return true;
}

void *GraphicsOpenGL::LoadTexture(SDL_Surface *p_Surface)
{
	// Local Variables
	unsigned int l_Texture;
	unsigned int l_TextureFormat;
	unsigned int l_NumOfColors;
	unsigned int *l_TextureHolder = new unsigned int;
	
	if (p_Surface->w & (p_Surface->w-1) || p_Surface->h & (p_Surface->h-1)) 
	{
		// NPOT texture...
	}
	
	l_NumOfColors = p_Surface->format->BytesPerPixel;
	if (l_NumOfColors == 4)
	{
		if (p_Surface->format->Rmask == 0x000000ff)
		{
			l_TextureFormat = GL_RGBA;
		}
		else
		{
			l_TextureFormat = GL_BGRA;
		}
	} 
	else 
	{
		if (l_NumOfColors == 3)
		{
			if (p_Surface->format->Rmask == 0x000000ff)
			{
				l_TextureFormat = GL_RGB;
			}
			else
			{
				l_TextureFormat = GL_BGR;
			}
		}
		else 
		{
			// Warn about palleted textures.
			fprintf(stderr,"Warning: Pallete textures encountered.\nThis is not recommended for use in modern OpenGL.\n");
		}
	}
	
	// Generate a new texture
	glGenTextures( 1, &l_Texture );
	
	// Tell OpenGL to use texture l_Texture
	glBindTexture( GL_TEXTURE_2D, l_Texture );
	
	// Set the filters used when rendering the texture.
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	
	// Load the texture data from p_Surface.
	glTexImage2D(GL_TEXTURE_2D, 0, l_NumOfColors, p_Surface->w, p_Surface->h, 0, l_TextureFormat, GL_UNSIGNED_BYTE, p_Surface->pixels);
	
	// Discard p_Surface if we still have it.
	if ( p_Surface ) 
	{
		SDL_FreeSurface( p_Surface );
	}
	
	// Set the texture ID to a pointer and return it.
	*l_TextureHolder = l_Texture;
	return l_TextureHolder;
}

void GraphicsOpenGL::DrawRect(void *p_Texture, TextureCoord p_TextureCoordinates, RectangleShape p_DrawRect, float p_Color[4])
{
	
	// Local Variables
	unsigned int l_Texture = *((unsigned int *)p_Texture);
	
	// Tell OpenGL to use texture l_Texture
	glBindTexture( GL_TEXTURE_2D, l_Texture );
	
	// Set the color tinting
	glColor4f(p_Color[0],p_Color[1],p_Color[2],p_Color[3]);
	
	// Do the actual drawing.
	glBegin(GL_QUADS);
	{
		glTexCoord2f(p_TextureCoordinates.X1, p_TextureCoordinates.Y1); 
		glVertex2f(p_DrawRect.X, p_DrawRect.Y);
		
		glTexCoord2f(p_TextureCoordinates.X2, p_TextureCoordinates.Y1); 
		glVertex2f(p_DrawRect.X+p_DrawRect.W, p_DrawRect.Y);
		
		glTexCoord2f(p_TextureCoordinates.X2, p_TextureCoordinates.Y2); 
		glVertex2f(p_DrawRect.X+p_DrawRect.W, p_DrawRect.Y+p_DrawRect.H);
		
		glTexCoord2f(p_TextureCoordinates.X1, p_TextureCoordinates.Y2); 
		glVertex2f(p_DrawRect.X, p_DrawRect.Y+p_DrawRect.H);
	}
	glEnd();
}

void GraphicsOpenGL::DrawText(wchar_t *p_String, bool p_IsLarge, Position p_Point, Orientation p_Orientation)
{
	glPushMatrix();
	{
		FTGLTextureFont *font = p_IsLarge?fontlarge:fontsmall;
		
		// Stores the bounding box of the string.
		float l_X1, l_X2, l_Y1, l_Y2, l_Z1, l_Z2;
		
		// This will enter in the bounding box for us.
		font->BBox(p_String, l_X1, l_Y1, l_Z1, l_X2, l_Y2, l_Z2);
		
		// This makes the top-left of the text to be the current position.
		glTranslatef(p_Point.X,p_Point.Y+(l_Y2),0);
		
		switch(p_Orientation)
		{
			case FONT_LEFT:
				break;
				
			case FONT_CENTER:
				glTranslatef(-(l_X2/2),0,0);
				break;
				
			case FONT_RIGHT:
				glTranslatef(-l_X2,0,0);
				break;
		}
		
		glScalef(1,-1,1);
		font->Render(p_String);
	}
	glPopMatrix();
}

void GraphicsOpenGL::Update()
{
	SDL_GL_SwapBuffers();
	glClear(GL_COLOR_BUFFER_BIT);
}
		
void GraphicsOpenGL::Close()
{
	SDL_Quit();
}

